# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject, DataTools
from tools.excel import ExcelTools, ExcelNavigator
from tools.text import PrettyText
from tools.parallel import Stoppable

from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher
from pathlib import Path
from tools.multisheet import MultiSheet


# -------------------------------------------------------------
# Converter: PANGAEA
# -------------------------------------------------------------

class ConverterPangaea(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):

        # Admin
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        # Daten
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()

        # Information Excel PANGAEA Sheet
        self.__sExcel_sheetname_data = "Raw data"
        self.__sExcel_sheetname_meta = "Data metadata description"
        self.__sSql_IndividualMeasureMeio_tableAttributeUsed4sorting = "INDMEA_MEIO_ID"
        self.__sSql_IndividualMeasureMacro_tableAttributeUsed4sorting = "INDMEA_MACRO_ID"
        self.__sSql_IndividualLabel_name = "Individual label"
        self.__sSql_Event_name = "Event"
        self.__lstMap_PANGAEA_CB_parts = ["branch", "leaf-biota", "leaf-indmeasure-meio", "leaf-indmeasure-macro", "leaf-sediment", "leaf-biogenic"]

        # Mapping
        self.__odicMap_PANGAEA_CB = OrderedDict()  # Name short: branch/leaf, Name long, Unit, DataType, Comment, PANGAEA code, CB Sheet, CB Col
        self.__dicPANGAEA_CB_idx2name = {}
        self.__dicPANGAEA_CB_name2idx = {}

        # SQL
        self.__sqlPANGAEA = \
            """
            SELECT * FROM (
            SELECT DISTINCT * FROM (
            -- BIOTA
            SELECT 
            CONCAT(cru.name, '_', sta.name, '-', sam.name, '.', subsmp.name, '.', sub.name) AS "Event",
            dat.name AS "Project", 
            cru.name AS "Campaign", 
            sta.name AS "Station", 
            sam.name AS "Sample", 
            subsmp.name AS "Subsample", 
            sub.name AS "Subset", 
            shp.name AS "Vessel",
            per_dat.full_name AS "Contact person",
            dat.reference_person AS "Reference person",
            sta.location AS "Station area", 
            sam.date_start AS "Date", 
            sam.time_start AS "Time", 
            trunc(sam.lat_start::numeric, sam.precision::integer) as "Latitude",
            trunc(sam.lon_start::numeric, sam.precision::integer) as "Longitude",    
            sam.precision AS "Precision", 
            sam.depth_start AS "Water depth", 
            sam.sampled_area AS "Sampled area", 
            sam.reference_area AS "Reference area", 
            sam.sampled_vol AS "Sampled volume", 
            sam.reference_vol AS "Reference volume", 
            sam.sampling_speed AS "Sampling speed", 
            sam.sampling_distance AS "Sampling distance", 
            subsmp.ratio AS "Subsample ratio", 
            gea.category AS "Gear category",
            gea.name AS "Gear",
            tcov.name AS "Taxonomic coverage",
            sub.taxres AS "Taxonomic resolution",
            tax.aid AS "AphiaID", 
            tax.scientificname AS "Scientific name", 
            tax.url AS "Taxon URL", 
            bio.presence_absence AS "Presence/Absence", 
            bio.number AS "Count", 
            bio.abundance AS "Abundance", 
            bio.biomass_wet AS "Biomass wet", 
            bio.biomass_dry AS "Biomass dry", 
            bio.biomass_afdm AS "Biomass AFDW", 
            lif.name AS "Lifestage", 
            bio.sieve AS "Sieve fraction", 
            bio.subtotal_factor AS "Sub total factor", 
            bio.specification AS "Specification",
            bio.remark AS "Taxon remark",
            bio.name AS "Individual label", 
            null::integer AS "INDMEA_MEIO_ID",
            null::text AS "Individual measurement meiofauna sex",
            null::float AS "Individual measurement meiofauna length total",
            null::float AS "Individual measurement meiofauna width",
            null::text AS "Individual measurement meiofauna remark",
            null::integer AS "INDMEA_MACRO_ID",
            null::text AS "Individual measurement macrofauna sex",
            null::float AS "Individual measurement macrofauna length total",
            null::float AS "Individual measurement macrofauna length std",
            null::float AS "Individual measurement macrofauna mass total",
            null::float AS "Individual measurement macrofauna mass slaughter",
            null::text AS "Individual measurement macrofauna remark",
            null::float AS "Sediment grain size 1",
            null::float AS "Sediment mesh size 1",
            null::float AS "Sediment grain size 2",
            null::float AS "Sediment mesh size 2",
            null::float AS "Sediment grain size 3",
            null::float AS "Sediment mesh size 3",
            null::float AS "Sediment grain size 4",
            null::float AS "Sediment mesh size 4",
            null::float AS "Sediment grain size 5",
            null::float AS "Sediment mesh size 5",
            null::float AS "Sediment grain size 6",
            null::float AS "Sediment mesh size 6",
            null::float AS "Sediment grain size 7",
            null::float AS "Sediment mesh size 7",
            null::float AS "Sediment grain size 8",
            null::float AS "Sediment mesh size 8",
            null::float AS "Sediment grain size 9",
            null::float AS "Sediment mesh size 9",
            null::float AS "Sediment sample weight",
            null::float AS "Sediment median",
            null::float AS "Sediment GSD",
            null::text AS "Sediment grain size method",
            null::float AS "Sediment sed temp",
            null::float AS "Sediment OxyLayer",
            null::float AS "Sediment Silt",
            null::float AS "Sediment SandVF",
            null::float AS "Sediment SandF",
            null::float AS "Sediment SandM",
            null::float AS "Sediment SandC",
            null::float AS "Sediment SandVC",
            null::float AS "Sediment gravel",
            null::float AS "Sediment stones",
            null::float AS "Sediment Schill",
            null::float AS "Sediment Klei",
            null::text AS "Sediment remark",
            null::float AS "BiogenicCompound C org",
            null::float AS "BiogenicCompound C total",
            null::float AS "BiogenicCompound loss on ignition",
            null::float AS "BiogenicCompound nitrogen",
            null::float AS "BiogenicCompound sulfur",
            null::float  AS "BiogenicCompound Chl A",
            null::text AS "BiogenicCompound remark"
            FROM <DBSCHEMA>.subset sub
            JOIN <DBSCHEMA>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <DBSCHEMA>.sample sam ON sam.id = subsmp.sample_id
            JOIN <DBSCHEMA>.station sta ON sta.id = sam.station_id
            JOIN <DBSCHEMA>.cruise cru ON cru.id = sta.cruise_id
            JOIN <DBSCHEMA>.biota bio ON sub.id = bio.subset_id
            JOIN <DBSCHEMA>.dataset dat ON dat.id = sub.dataset_id
            JOIN <DBSCHEMA>.gear gea ON gea.id = sam.gear_id
            JOIN <DBSCHEMA>.crs crs ON crs.id = cru.crs_id
            JOIN <DBSCHEMA>.taxon tax ON bio.taxon_id=tax.id 
            LEFT JOIN <DBSCHEMA>.taxcover tcov ON sub.taxcover_id=tcov.id 
            LEFT JOIN <DBSCHEMA>.ship shp ON shp.id = cru.ship_id
            LEFT JOIN <DBSCHEMA>.person per_dat ON per_dat.id = dat.contact_person_id
            LEFT JOIN <DBSCHEMA>.lifestage lif ON bio.lifestage_id=lif.id 
            WHERE sub.id IN (SELECT id FROM selector) 
            UNION
            -- INDMEASURE MEIO
            SELECT 
            CONCAT(cru.name, '_', sta.name, '-', sam.name, '.', subsmp.name, '.', sub.name) AS "Event",
            dat.name AS "Project", 
            cru.name AS "Campaign", 
            sta.name AS "Station", 
            sam.name AS "Sample", 
            subsmp.name AS "Subsample", 
            sub.name AS "Subset", 
            shp.name AS "Vessel",
            per_dat.full_name AS "Contact person",
            dat.reference_person AS "Reference person",
            sta.location AS "Station area", 
            sam.date_start AS "Date", 
            sam.time_start AS "Time", 
            --crs.name AS "CRS",
            trunc(sam.lat_start::numeric, sam.precision::integer) as "Latitude",
            trunc(sam.lon_start::numeric, sam.precision::integer) as "Longitude",    
            sam.precision AS "Precision", 
            sam.depth_start AS "Water depth", 
            sam.sampled_area AS "Sampled area", 
            sam.reference_area AS "Reference area", 
            sam.sampled_vol AS "Sampled volume", 
            sam.reference_vol AS "Reference volume", 
            sam.sampling_speed AS "Sampling speed", 
            sam.sampling_distance AS "Sampling distance", 
            subsmp.ratio AS "Subsample ratio", 
            gea.category AS "Gear category",
            gea.name AS "Gear",
            tcov.name AS "Taxonomic coverage",
            sub.taxres AS "Taxonomic resolution",
            null::integer AS "AphiaID", 
            null::text AS "Scientific name", 
            null::text AS "Taxon URL", 
            null::integer AS "Presence/Absence", 
            null::numeric AS "Count", 
            null::numeric AS "Abundance", 
            null::float AS "Biomass wet", 
            null::float AS "Biomass dry", 
            null::float AS "Biomass AFDW", 
            null::text AS "Lifestage", 
            null::integer AS "Sieve fraction", 
            null::text AS "Sub total factor", 
            null::text AS "Specification",
            null::text AS "Taxon remark",
            bio.name AS "Individual label", 
            mea.id AS "INDMEA_MEIO_ID",
            mea.sex AS "Individual measurement meiofauna sex",
            mea.length_total AS "Individual measurement meiofauna length total",
            mea.width AS "Individual measurement meiofauna width",
            mea.remark AS "Individual measurement meiofauna remark",
            null::integer AS "INDMEA_MACRO_ID",
            null::text AS "Individual measurement macrofauna sex",
            null::float AS "Individual measurement macrofauna length total",
            null::float AS "Individual measurement macrofauna length std",
            null::float AS "Individual measurement macrofauna mass total",
            null::float AS "Individual measurement macrofauna mass slaughter",
            null::text AS "Individual measurement macrofauna remark",
            null::float AS "Sediment grain size 1",
            null::float AS "Sediment mesh size 1",
            null::float AS "Sediment grain size 2",
            null::float AS "Sediment mesh size 2",
            null::float AS "Sediment grain size 3",
            null::float AS "Sediment mesh size 3",
            null::float AS "Sediment grain size 4",
            null::float AS "Sediment mesh size 4",
            null::float AS "Sediment grain size 5",
            null::float AS "Sediment mesh size 5",
            null::float AS "Sediment grain size 6",
            null::float AS "Sediment mesh size 6",
            null::float AS "Sediment grain size 7",
            null::float AS "Sediment mesh size 7",
            null::float AS "Sediment grain size 8",
            null::float AS "Sediment mesh size 8",
            null::float AS "Sediment grain size 9",
            null::float AS "Sediment mesh size 9",
            null::float AS "Sediment sample weight",
            null::float AS "Sediment median",
            null::float AS "Sediment GSD",
            null::text AS "Sediment grain size method",
            null::float AS "Sediment sed temp",
            null::float AS "Sediment OxyLayer",
            null::float AS "Sediment Silt",
            null::float AS "Sediment SandVF",
            null::float AS "Sediment SandF",
            null::float AS "Sediment SandM",
            null::float AS "Sediment SandC",
            null::float AS "Sediment SandVC",
            null::float AS "Sediment gravel",
            null::float AS "Sediment stones",
            null::float AS "Sediment Schill",
            null::float AS "Sediment Klei",
            null::text AS "Sediment remark",
            null::float AS "BiogenicCompound C org",
            null::float AS "BiogenicCompound C total",
            null::float AS "BiogenicCompound loss on ignition",
            null::float AS "BiogenicCompound nitrogen",
            null::float AS "BiogenicCompound sulfur",
            null::float  AS "BiogenicCompound Chl A",
            null::text AS "BiogenicCompound remark"
            FROM <DBSCHEMA>.subset sub
            JOIN <DBSCHEMA>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <DBSCHEMA>.sample sam ON sam.id = subsmp.sample_id
            JOIN <DBSCHEMA>.station sta ON sta.id = sam.station_id
            JOIN <DBSCHEMA>.cruise cru ON cru.id = sta.cruise_id
            JOIN <DBSCHEMA>.biota bio ON sub.id = bio.subset_id
            JOIN <DBSCHEMA>.imeasure_meio mea ON bio.id = mea.biota_id
            JOIN <DBSCHEMA>.dataset dat ON dat.id = sub.dataset_id
            JOIN <DBSCHEMA>.gear gea ON gea.id = sam.gear_id
            JOIN <DBSCHEMA>.crs crs ON crs.id = cru.crs_id
            JOIN <DBSCHEMA>.taxon tax ON bio.taxon_id=tax.id 
            LEFT JOIN <DBSCHEMA>.taxcover tcov ON sub.taxcover_id=tcov.id 
            LEFT JOIN <DBSCHEMA>.ship shp ON shp.id = cru.ship_id
            LEFT JOIN <DBSCHEMA>.person per_dat ON per_dat.id = dat.contact_person_id
            WHERE sub.id IN (SELECT id FROM selector) 
            UNION	
            -- INDMEASURE MACRO
            SELECT 
            CONCAT(cru.name, '_', sta.name, '-', sam.name, '.', subsmp.name, '.', sub.name) AS "Event",
            dat.name AS "Project", 
            cru.name AS "Campaign", 
            sta.name AS "Station", 
            sam.name AS "Sample", 
            subsmp.name AS "Subsample", 
            sub.name AS "Subset", 
            shp.name AS "Vessel",
            per_dat.full_name AS "Contact person",
            dat.reference_person AS "Reference person",
            sta.location AS "Station area", 
            sam.date_start AS "Date", 
            sam.time_start AS "Time", 
            trunc(sam.lat_start::numeric, sam.precision::integer) as "Latitude",
            trunc(sam.lon_start::numeric, sam.precision::integer) as "Longitude",    
            sam.precision AS "Precision", 
            sam.depth_start AS "Water depth", 
            sam.sampled_area AS "Sampled area", 
            sam.reference_area AS "Reference area", 
            sam.sampled_vol AS "Sampled volume", 
            sam.reference_vol AS "Reference volume", 
            sam.sampling_speed AS "Sampling speed", 
            sam.sampling_distance AS "Sampling distance", 
            subsmp.ratio AS "Subsample ratio", 
            gea.category AS "Gear category",
            gea.name AS "Gear",
            tcov.name AS "Taxonomic coverage",
            sub.taxres AS "Taxonomic resolution",
            null::integer AS "AphiaID", 
            null::text AS "Scientific name", 
            null::text AS "Taxon URL", 
            null::integer AS "Presence/Absence", 
            null::numeric AS "Count", 
            null::numeric AS "Abundance", 
            null::float AS "Biomass wet", 
            null::float AS "Biomass dry", 
            null::float AS "Biomass AFDW", 
            null::text AS "Lifestage", 
            null::integer AS "Sieve fraction", 
            null::text AS "Sub total factor", 
            null::text AS "Specification",
            null::text AS "Taxon remark",
            bio.name AS "Individual label", 
            null::integer AS "INDMEA_MEIO_ID",
            null::text AS "Individual measurement meiofauna sex",
            null::float AS "Individual measurement meiofauna length total",
            null::float AS "Individual measurement meiofauna width",
            null::text AS "Individual measurement meiofauna remark",
            mea.id AS "INDMEA_MACRO_ID",
            mea.sex AS "Individual measurement macrofauna sex",
            mea.length_total AS "Individual measurement macrofauna length total",
            mea.length_st AS "Individual measurement macrofauna length std",
            mea.mass_total AS "Individual measurement macrofauna mass total",
            mea.mass_sh AS "Individual measurement macrofauna mass slaughter",
            mea.remark AS "Individual measurement macrofauna remark",
            null::float AS "Sediment grain size 1",
            null::float AS "Sediment mesh size 1",
            null::float AS "Sediment grain size 2",
            null::float AS "Sediment mesh size 2",
            null::float AS "Sediment grain size 3",
            null::float AS "Sediment mesh size 3",
            null::float AS "Sediment grain size 4",
            null::float AS "Sediment mesh size 4",
            null::float AS "Sediment grain size 5",
            null::float AS "Sediment mesh size 5",
            null::float AS "Sediment grain size 6",
            null::float AS "Sediment mesh size 6",
            null::float AS "Sediment grain size 7",
            null::float AS "Sediment mesh size 7",
            null::float AS "Sediment grain size 8",
            null::float AS "Sediment mesh size 8",
            null::float AS "Sediment grain size 9",
            null::float AS "Sediment mesh size 9",
            null::float AS "Sediment sample weight",
            null::float AS "Sediment median",
            null::float AS "Sediment GSD",
            null::text AS "Sediment grain size method",
            null::float AS "Sediment sed temp",
            null::float AS "Sediment OxyLayer",
            null::float AS "Sediment Silt",
            null::float AS "Sediment SandVF",
            null::float AS "Sediment SandF",
            null::float AS "Sediment SandM",
            null::float AS "Sediment SandC",
            null::float AS "Sediment SandVC",
            null::float AS "Sediment gravel",
            null::float AS "Sediment stones",
            null::float AS "Sediment Schill",
            null::float AS "Sediment Klei",
            null::text AS "Sediment remark",
            null::float AS "BiogenicCompound C org",
            null::float AS "BiogenicCompound C total",
            null::float AS "BiogenicCompound loss on ignition",
            null::float AS "BiogenicCompound nitrogen",
            null::float AS "BiogenicCompound sulfur",
            null::float  AS "BiogenicCompound Chl A",
            null::text AS "BiogenicCompound remark"
            FROM <DBSCHEMA>.subset sub
            JOIN <DBSCHEMA>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <DBSCHEMA>.sample sam ON sam.id = subsmp.sample_id
            JOIN <DBSCHEMA>.station sta ON sta.id = sam.station_id
            JOIN <DBSCHEMA>.cruise cru ON cru.id = sta.cruise_id
            JOIN <DBSCHEMA>.biota bio ON sub.id = bio.subset_id
            JOIN <DBSCHEMA>.imeasure_macro mea ON bio.id = mea.biota_id
            JOIN <DBSCHEMA>.dataset dat ON dat.id = sub.dataset_id
            JOIN <DBSCHEMA>.gear gea ON gea.id = sam.gear_id
            JOIN <DBSCHEMA>.crs crs ON crs.id = cru.crs_id
            JOIN <DBSCHEMA>.taxon tax ON bio.taxon_id=tax.id 
            LEFT JOIN <DBSCHEMA>.taxcover tcov ON sub.taxcover_id=tcov.id 
            LEFT JOIN <DBSCHEMA>.ship shp ON shp.id = cru.ship_id
            LEFT JOIN <DBSCHEMA>.person per_dat ON per_dat.id = dat.contact_person_id
            WHERE sub.id IN (SELECT id FROM selector) 
            UNION	
            -- SEDIMENT
            SELECT 
            CONCAT(cru.name, '_', sta.name, '-', sam.name, '.', subsmp.name, '.', sub.name) AS "Event",
            dat.name AS "Project", 
            cru.name AS "Campaign", 
            sta.name AS "Station", 
            sam.name AS "Sample", 
            subsmp.name AS "Subsample", 
            sub.name AS "Subset", 
            shp.name AS "Vessel",
            per_dat.full_name AS "Contact person",
            dat.reference_person AS "Reference person",
            sta.location AS "Station area", 
            sam.date_start AS "Date", 
            sam.time_start AS "Time", 
            trunc(sam.lat_start::numeric, sam.precision::integer) as "Latitude",
            trunc(sam.lon_start::numeric, sam.precision::integer) as "Longitude",    
            sam.precision AS "Precision", 
            sam.depth_start AS "Water depth", 
            sam.sampled_area AS "Sampled area", 
            sam.reference_area AS "Reference area", 
            sam.sampled_vol AS "Sampled volume", 
            sam.reference_vol AS "Reference volume", 
            sam.sampling_speed AS "Sampling speed", 
            sam.sampling_distance AS "Sampling distance", 
            subsmp.ratio AS "Subsample ratio", 
            gea.category AS "Gear category",
            gea.name AS "Gear",
            tcov.name AS "Taxonomic coverage",
            sub.taxres AS "Taxonomic resolution",
            null::integer AS "AphiaID", 
            null::text AS "Scientific name", 
            null::text AS "Taxon URL", 
            null::integer AS "Presence/Absence", 
            null::numeric AS "Count", 
            null::numeric AS "Abundance", 
            null::float AS "Biomass wet", 
            null::float AS "Biomass dry", 
            null::float AS "Biomass AFDW", 
            null::text AS "Lifestage", 
            null::integer AS "Sieve fraction", 
            null::text AS "Sub total factor", 
            null::text AS "Specification",
            null::text AS "Taxon remark",
            null::text AS "Individual label", 
            null::integer AS "INDMEA_MEIO_ID",
            null::text AS "Individual measurement meiofauna sex",
            null::float AS "Individual measurement meiofauna length total",
            null::float AS "Individual measurement meiofauna width",
            null::text AS "Individual measurement meiofauna remark",
            null::integer AS "INDMEA_MACRO_ID",
            null::text AS "Individual measurement macrofauna sex",
            null::float AS "Individual measurement macrofauna length total",
            null::float AS "Individual measurement macrofauna length std",
            null::float AS "Individual measurement macrofauna mass total",
            null::float AS "Individual measurement macrofauna mass slaughter",
            null::text AS "Individual measurement macrofauna remark",
            sed.residue1 AS "Sediment grain size 1",
            sed.meshsize1 AS "Sediment mesh size 1",
            sed.residue2 AS "Sediment grain size 2",
            sed.meshsize2 AS "Sediment mesh size 2",
            sed.residue3 AS "Sediment grain size 3",
            sed.meshsize3 AS "Sediment mesh size 3",
            sed.residue4 AS "Sediment grain size 4",
            sed.meshsize4 AS "Sediment mesh size 4",
            sed.residue5 AS "Sediment grain size 5",
            sed.meshsize5 AS "Sediment mesh size 5",
            sed.residue6 AS "Sediment grain size 6",
            sed.meshsize6 AS "Sediment mesh size 6",
            sed.residue7 AS "Sediment grain size 7",
            sed.meshsize7 AS "Sediment mesh size 7",
            sed.residue8 AS "Sediment grain size 8",
            sed.meshsize8 AS "Sediment mesh size 8",
            sed.residue9 AS "Sediment grain size 9",
            sed.meshsize9 AS "Sediment mesh size 9",
            sed.weight AS "Sediment sample weight",
            sed.median AS "Sediment median",
            sed.gsd AS "Sediment GSD",
            sed.method_grainsize AS "Sediment grain size method",
            sed.temperature AS "Sediment sed temp",
            sed.oxygen_pd AS "Sediment OxyLayer",
            sed.fuz_silt AS "Sediment Silt",
            sed.fuz_vfsand AS "Sediment SandVF",
            sed.fuz_fsand AS "Sediment SandF",
            sed.fuz_msand AS "Sediment SandM",
            sed.fuz_csand AS "Sediment SandC",
            sed.fuz_vcsand AS "Sediment SandVC",
            sed.fuz_gravel AS "Sediment gravel",
            sed.fuz_stones AS "Sediment stones",
            sed.fuz_shell AS "Sediment Schill",
            sed.fuz_clay AS "Sediment Klei",
            sed.remark AS "Sediment remark",
            null::float AS "BiogenicCompound C org",
            null::float AS "BiogenicCompound C total",
            null::float AS "BiogenicCompound loss on ignition",
            null::float AS "BiogenicCompound nitrogen",
            null::float AS "BiogenicCompound sulfur",
            null::float AS "BiogenicCompound Chl A",
            null::text AS "BiogenicCompound remark"
            FROM <DBSCHEMA>.subset sub
            JOIN <DBSCHEMA>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <DBSCHEMA>.sample sam ON sam.id = subsmp.sample_id
            JOIN <DBSCHEMA>.station sta ON sta.id = sam.station_id
            JOIN <DBSCHEMA>.cruise cru ON cru.id = sta.cruise_id
            JOIN <DBSCHEMA>.sediment sed ON sub.id = sed.subset_id	
            JOIN <DBSCHEMA>.dataset dat ON dat.id = sub.dataset_id
            JOIN <DBSCHEMA>.gear gea ON gea.id = sam.gear_id
            JOIN <DBSCHEMA>.crs crs ON crs.id = cru.crs_id
            LEFT JOIN <DBSCHEMA>.taxcover tcov ON sub.taxcover_id=tcov.id 
            LEFT JOIN <DBSCHEMA>.ship shp ON shp.id = cru.ship_id
            LEFT JOIN <DBSCHEMA>.person per_dat ON per_dat.id = dat.contact_person_id
            --LEFT JOIN <DBSCHEMA>.lifestage lif ON bio.lifestage_id=lif.id 
            WHERE sub.id IN (SELECT id FROM selector) 
            UNION
            -- BIOGENIC COMPOUND
            SELECT 
            CONCAT(cru.name, '_', sta.name, '-', sam.name, '.', subsmp.name, '.', sub.name) AS "Event",
            dat.name AS "Project", 
            cru.name AS "Campaign", 
            sta.name AS "Station", 
            sam.name AS "Sample", 
            subsmp.name AS "Subsample", 
            sub.name AS "Subset", 
            shp.name AS "Vessel",
            per_dat.full_name AS "Contact person",
            dat.reference_person AS "Reference person",
            sta.location AS "Station area", 
            sam.date_start AS "Date", 
            sam.time_start AS "Time", 
            trunc(sam.lat_start::numeric, sam.precision::integer) as "Latitude",
            trunc(sam.lon_start::numeric, sam.precision::integer) as "Longitude",    
            sam.precision AS "Precision", 
            sam.depth_start AS "Water depth", 
            sam.sampled_area AS "Sampled area", 
            sam.reference_area AS "Reference area", 
            sam.sampled_vol AS "Sampled volume", 
            sam.reference_vol AS "Reference volume", 
            sam.sampling_speed AS "Sampling speed", 
            sam.sampling_distance AS "Sampling distance", 
            subsmp.ratio AS "Subsample ratio", 
            gea.category AS "Gear category",
            gea.name AS "Gear",
            tcov.name AS "Taxonomic coverage",
            sub.taxres AS "Taxonomic resolution",
            null::integer AS "AphiaID", 
            null::text AS "Scientific name", 
            null::text AS "Taxon URL", 
            null::integer AS "Presence/Absence", 
            null::numeric AS "Count", 
            null::numeric AS "Abundance", 
            null::float AS "Biomass wet", 
            null::float AS "Biomass dry", 
            null::float AS "Biomass AFDW", 
            null::text AS "Lifestage", 
            null::integer AS "Sieve fraction", 
            null::text AS "Sub total factor", 
            null::text AS "Specification",
            null::text AS "Taxon remark",
            null::text AS "Individual label", 
            null::integer AS "INDMEA_MEIO_ID",
            null::text AS "Individual measurement meiofauna sex",
            null::float AS "Individual measurement meiofauna length total",
            null::float AS "Individual measurement meiofauna width",
            null::text AS "Individual measurement meiofauna remark",
            null::integer AS "INDMEA_MACRO_ID",
            null::text AS "Individual measurement macrofauna sex",
            null::float AS "Individual measurement macrofauna length total",
            null::float AS "Individual measurement macrofauna length std",
            null::float AS "Individual measurement macrofauna mass total",
            null::float AS "Individual measurement macrofauna mass slaughter",
            null::text AS "Individual measurement macrofauna remark",
            null::float AS "Sediment grain size 1",
            null::float AS "Sediment mesh size 1",
            null::float AS "Sediment grain size 2",
            null::float AS "Sediment mesh size 2",
            null::float AS "Sediment grain size 3",
            null::float AS "Sediment mesh size 3",
            null::float AS "Sediment grain size 4",
            null::float AS "Sediment mesh size 4",
            null::float AS "Sediment grain size 5",
            null::float AS "Sediment mesh size 5",
            null::float AS "Sediment grain size 6",
            null::float AS "Sediment mesh size 6",
            null::float AS "Sediment grain size 7",
            null::float AS "Sediment mesh size 7",
            null::float AS "Sediment grain size 8",
            null::float AS "Sediment mesh size 8",
            null::float AS "Sediment grain size 9",
            null::float AS "Sediment mesh size 9",
            null::float AS "Sediment sample weight",
            null::float AS "Sediment median",
            null::float AS "Sediment GSD",
            null::text AS "Sediment grain size method",
            null::float AS "Sediment sed temp",
            null::float AS "Sediment OxyLayer",
            null::float AS "Sediment Silt",
            null::float AS "Sediment SandVF",
            null::float AS "Sediment SandF",
            null::float AS "Sediment SandM",
            null::float AS "Sediment SandC",
            null::float AS "Sediment SandVC",
            null::float AS "Sediment gravel",
            null::float AS "Sediment stones",
            null::float AS "Sediment Schill",
            null::float AS "Sediment Klei",
            null::text AS "Sediment remark",
            gen.c_org AS "BiogenicCompound C org",
            gen.c_total AS "BiogenicCompound C total",
            gen.loss_on_ignition AS "BiogenicCompound loss on ignition",
            gen.nitrogen AS "BiogenicCompound nitrogen",
            gen.sulfur AS "BiogenicCompound sulfur",
            gen.chl_a AS "BiogenicCompound Chl A",
            gen.remark AS "BiogenicCompound remark"
            FROM <DBSCHEMA>.subset sub
            JOIN <DBSCHEMA>.subsample subsmp ON subsmp.id = sub.subsample_id
            JOIN <DBSCHEMA>.sample sam ON sam.id = subsmp.sample_id
            JOIN <DBSCHEMA>.station sta ON sta.id = sam.station_id
            JOIN <DBSCHEMA>.cruise cru ON cru.id = sta.cruise_id
            JOIN <DBSCHEMA>.biogenic gen ON sub.id = gen.subset_id	
            JOIN <DBSCHEMA>.dataset dat ON dat.id = sub.dataset_id
            JOIN <DBSCHEMA>.gear gea ON gea.id = sam.gear_id
            JOIN <DBSCHEMA>.crs crs ON crs.id = cru.crs_id
            LEFT JOIN <DBSCHEMA>.taxcover tcov ON sub.taxcover_id=tcov.id 
            LEFT JOIN <DBSCHEMA>.ship shp ON shp.id = cru.ship_id
            LEFT JOIN <DBSCHEMA>.person per_dat ON per_dat.id = dat.contact_person_id
            WHERE sub.id IN (SELECT id FROM selector) 
            ) t1
            ) t2 
            ORDER BY length ("Event") ASC, "Event" ASC, length ("Project") ASC, "Project" ASC, length ("Campaign") ASC, "Campaign" ASC, length ("Station") ASC, "Station" ASC, length ("Sample") ASC, "Sample" ASC, length ("Subsample") ASC, "Subsample" ASC, length ("Subset") ASC, "Subset" ASC, length ("Individual label") ASC, "Individual label" ASC, "INDMEA_MEIO_ID" DESC, "INDMEA_MACRO_ID" DESC
            """

    # Initialisation
    def init(self):

        # Helper
        # Clean Whitespaces
        def clean(sTxt):
            if sTxt is None:
                return None
            sTxt = str(sTxt)
            sTxt_clean = sTxt.strip()
            if sTxt != sTxt_clean:
                if res is not None and iRow is not None:
                    res.addInfo("Whitespace at start and end found and removed", add=str(sTxt),
                                kat="Initialisation: Cleaning Definiton Data", sys="Converter: PANGAEA", sheet="Def",
                                row=iRow)
                else:
                    print(
                        "Converter: MARLIN: Cleaning Definiton Data: Whitespace at start/end removed: " + str(sTxt))
            return sTxt_clean

        # ---

        res = Result("Mapping CB - PANGAEA")

        # Erzeuge Mapping PANGAEA-CB
        sMap = os.path.dirname(os.path.realpath(__file__)) + "/mapper_convert_pangaea.csv"
        try:
            # Einlesen der CSV-Datei
            with open(sMap, newline='', encoding='utf-8') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                # Zeilen
                for iRow, lstZeile in enumerate(csvReader):
                    if iRow > 0:
                        # Hole Werte
                        sP_name_short = clean(lstZeile[0])
                        sP_name_long = clean(lstZeile[1])
                        sP_code = clean(lstZeile[2])
                        sCB_part = clean(lstZeile[3])
                        sCB_sheet = clean(lstZeile[4])
                        sCB_col = clean(lstZeile[5])
                        sOverride_unit = clean(lstZeile[6])
                        sOverride_datatype = clean(lstZeile[7])
                        sOverride_comment = clean(lstZeile[8])
                        sGold = clean(lstZeile[9])
                        # Screibe in Map
                        self.__odicMap_PANGAEA_CB[sP_name_short] = (
                            sCB_part, sP_name_long, sOverride_unit, sOverride_datatype, sOverride_comment, sP_code,
                            sCB_sheet, sCB_col, sGold)
        except Exception as ex:
            res.addErr("Unexpected problem: " + str(ex), kat="Initialisation", sys="Converter: PANGAEA", sheet="Def",
                       row=iRow)

        # Erzeuge Zusatzinfo: Index<->Name
        for iIdx, key in enumerate(self.__odicMap_PANGAEA_CB.keys()):
            self.__dicPANGAEA_CB_idx2name[iIdx] = key
            self.__dicPANGAEA_CB_name2idx[key] = iIdx

        # Working
        self.setWorking(True)
        self.setReady(True)

        # Antwort
        if res.isWarn():
            print(res.getWarn())

        if res.isErr():
            print(res.getErr())
            self.setReady(False)

        return res

    # Admin
    def map_getIndex(self, sName_short):
        return self.__dicPANGAEA_CB_name2idx.get(sName_short)

    def map_getName(self, sIndex):
        return self.__dicPANGAEA_CB_idx2name.get(sIndex)

    def map_getPart(self, sIndex):
        sName_short = self.__dicPANGAEA_CB_idx2name.get(sIndex)
        (sCB_part, sP_name_long, sOverride_unit, sOverride_datatype, sOverride_comment, sP_code, sCB_sheet, sCB_col,
         sGold) = self.__odicMap_PANGAEA_CB.get(sName_short)
        return sCB_part

    # Info
    def info(self, bHtml):
        sOut = "************************************\n"
        sOut += "Export: CB -> PANGAEA\n"
        sOut += "************************************\n\n"
        sOut += "Export data to PANGAEA format\n"
        return sOut

    # Reset
    def reset(self):
        pass

    # Converter PANGAEA
    def convert_export(self, sSelector_sql):
        """
        Args:
            sSelector_sql (str): Selector Sql
        Kwargs:
            -
        Return:
            Result
                data (odic)
                stat (odic)
                log (str)
        """

        # Helper
        # Clean DB Connectio
        def cleanUp():
            # DB
            try:
                cursor.close()
                conn.close()
            except:
                pass

        # Create Uniq Event-IMeasure-ID
        def creatEventImeasureId(row):
            sUniq_Event_Indmeasurelabel = None
            try:
                lstRow = list(row)
                sIndmeasureLabel = lstRow[self.map_getIndex(self.__sSql_IndividualLabel_name)]
                sEvent = lstRow[self.map_getIndex(self.__sSql_Event_name)]
                if sEvent is not None and sIndmeasureLabel is not None:
                    sUniq_Event_Indmeasurelabel = sEvent + "-" + sIndmeasureLabel
            except:
                pass
            return sUniq_Event_Indmeasurelabel

        # Create PANGAEA style Header
        def createPANGAEA(sName_short, iPangaea, sComment, sUnit):
            sName_pangaeaStyle = ""
            if iPangaea is not None:
                sName_pangaeaStyle += iPangaea + " "
                if sComment:
                    sName_pangaeaStyle += "@" + sComment
            else:
                if sName_short:
                    sName_pangaeaStyle += sName_short + " "
                if sUnit:
                    sName_pangaeaStyle += "[" + sUnit + "] "
                if sComment:
                    sName_pangaeaStyle += "@" + sComment
            return sName_pangaeaStyle if sName_pangaeaStyle != "" else None

        # ---

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.verbose
        workflow = self.__smartprinter.workflow
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status

        res = Result("Convert CB -> PANGAEA Export")

        # DB
        resConn = self.__model.subsystem_getDb().conn_getConnection()
        if resConn.isErr():
            cleanUp()
            return resConn
        conn = resConn.getData("connection")
        cursor = resConn.getData("cursor")

        # Hole SQL Befehle
        sDbSchema = self.__model.subsystem_getDb().account_active_getSchema()
        if sDbSchema is None:
            res.addErr("Creating SQL for PANGAEA Export", add="Schema is not known", kat="SQL", sys="PANGAEA Exporter")
            cleanUp()
            return res
        sPANGAEA_Sql = self.__sqlPANGAEA.replace("<DBSCHEMA>",
                                                 self.__model.subsystem_getDb().account_active_getSchema())
        sPANGAEA_Sql = PrettyText.clean_delWhitespaces_everyline(sPANGAEA_Sql)

        # LOG
        sLog = "--- SQL (PANGAEA flat pack) ---\n"
        sLog += sPANGAEA_Sql + "\n"

        # PANGAEA data
        # data (sql tabelle)
        lstPANGAEA_data_header_sql = []
        lstPANGAEA_data_sql = []

        # data (intern)
        lstPANGAEA_data_header = []
        lstPANGAEA_data_header_clean = []  # clean (leere spalten entfernt)
        lstPANGAEA_data = []
        lstPANGAEA_data_clean = []

        # meta data (intern)
        lstPANGAEA_meta_header = ["Name", "Short Name", "Unit", "Data type", "Comment", "PANGAEA code"]
        lstPANGAEA_meta = []
        lstPANGAEA_meta_clean = []  # clean (leere spalten entfernt)

        # *** Hole Daten aus Datenbank
        try:
            workflow(top="Converter CB->PANGAEA", sub="Asking data base", fmt=False)
            progress("on-pulse")

            # Selector View aufbauen
            cursor.execute(sSelector_sql)

            # Hole Daten
            cursor.execute(sPANGAEA_Sql)

            # Antwort
            lstPANGAEA_data_header_sql = [head[0] for head in cursor.description]
            lstPANGAEA_data_sql = cursor.fetchall()

        except Exception as ex:
            res.addErr("Getting data for PANGAEA Export", add=str(ex), kat="SQL", sys="PANGAEA Exporter")
        if res.isErr():
            return res

        # *** Erzeuge Rahmen und Beschreibung für interne Datenstruktur (Data.Header und Metadata.Rows)
        # Log
        lstPANGAEA_meta_log = []
        lstPANGAEA_meta_log_header = ["Name", "Short Name", "Unit", "Unit CB", "Data type", "Data type CB", "Comment",
                                      "Comment CB", "PANGAEA code", "Gold"]
        lstPANGAEA_meta_log.append(lstPANGAEA_meta_log_header)
        iGold_total = 0
        iGold_unset = 0
        for sP_name_short, (
                sCB_part, sP_name_long, sOverride_unit, sOverride_datatype, sOverride_comment, sP_code, sCB_sheet,
                sCB_col,
                sGold) in self.__odicMap_PANGAEA_CB.items():

            # Daten aus dem CB Data space
            sCB_name = None
            sCB_datatype = None
            sCB_unit = None
            sCB_comment = None

            # Finale Version
            sFinal_name_short = None
            sFinal_name_long = None
            sFinal_unit = None
            sFinal_datatype = None
            sFinal_comment = None
            sFinal_code = None

            # Check CB part
            if sCB_part not in self.__lstMap_PANGAEA_CB_parts:
                res.addErr("Unknown CB-PANGAEA mapping part", add=str(sCB_part), kat="Create EXCEL",
                           sys="PANGAEA Exporter")

            if sP_name_short in ("Event", self.__sSql_IndividualMeasureMacro_tableAttributeUsed4sorting, self.__sSql_IndividualMeasureMeio_tableAttributeUsed4sorting):
                # Spezialfaelle
                pass
            elif not self.__map.all_isSheetColExisting(sCB_sheet, sCB_col):
                # Gibt es die Kombination Sheet/Col in CB
                res.addErr("Unknown CB Column/Sheet Name", add=str(sCB_sheet) + "." + str(sCB_col), kat="Create EXCEL",
                           sys="PANGAEA Exporter")
            elif self.__map.all_isSheetIngest(sCB_sheet):
                # Hole CB infos fuer Ingest Sheet
                sCB_name = self.__map.ingest_getName(sCB_sheet, sCB_col)
                sCB_comment = self.__map.ingest_getComment(sCB_sheet, sCB_col)
                sCB_unit = self.__map.ingest_getUnit(sCB_sheet, sCB_col)
                sCB_datatype = self.__model.subsystem_getSchema().getDbSchemaInfo("ingest", sCB_sheet, sCB_col).getData(
                    "datatype")  # etwas tricky
                sCB_pangaea = self.__map.ingest_getCode_foreign(sCB_sheet, sCB_col)
            elif self.__map.all_isSheetLookup(sCB_sheet):
                # Hole CB infos fuer Lookup Sheet
                sCB_name = self.__map.lookup_getName(sCB_sheet, sCB_col)
                sCB_comment = self.__map.lookup_getComment(sCB_sheet, sCB_col)
                sCB_unit = self.__map.lookup_getUnit(sCB_sheet, sCB_col)
                sCB_datatype = self.__model.subsystem_getSchema().getDbSchemaInfo("lookup", sCB_sheet, sCB_col).getData(
                    "datatype")  # etwas tricky
                sCB_pangaea = self.__map.lookup_getCode_foreign(sCB_sheet, sCB_col)
            else:
                # Unbekannt -> Fehler
                res.addErr("Uuups this should never happen", add=str(sP_name_short), kat="Create EXCEL",
                           sys="PANGAEA Exporter")

            # --- Erzeuge Metadaten Header
            # Name Short
            sFinal_name_short = sP_name_short
            # Name long
            sFinal_name_long = sP_name_long
            if sP_name_long in [None, ""]:
                # fehlt? <- Name short
                sFinal_name_long = sP_name_short
            # Pangaea Code
            if sP_code not in [None, ""]:
                sFinal_code = sP_code

            # Ueberschreibe Daten wenn gewuenscht
            # benutze override: wenn nicht ""
            # benutze cb info: wenn override = "cb"
            # Unit
            if sOverride_unit == "cb":
                sFinal_unit = sCB_unit
            elif sOverride_unit not in (None, ""):
                sFinal_unit = sOverride_unit
            # Data type
            if sOverride_datatype == "cb":
                sFinal_datatype = sCB_datatype
            elif sOverride_datatype not in (None, ""):
                sFinal_datatype = sOverride_datatype
            # Comment
            if sOverride_comment == "cb":
                sFinal_comment = sCB_comment
            elif sOverride_comment not in (None, ""):
                sFinal_comment = sOverride_comment

            # Erzeuge Zeile: PANGAEA meta
            lstEntry = [sFinal_name_long, sFinal_name_short, sFinal_unit, sFinal_datatype, sFinal_comment, sFinal_code]
            lstPANGAEA_meta.append(lstEntry)

            # Log
            sOverride_unit = "-->" if sOverride_unit == "cb" else sOverride_unit
            sOverride_datatype = "-->" if sOverride_datatype == "cb" else sOverride_datatype
            sOverride_comment = "-->" if sOverride_comment == "cb" else sOverride_comment
            lstEntry_log = [sFinal_name_long, sFinal_name_short, sOverride_unit, sCB_unit, sOverride_datatype,
                            sCB_datatype, sOverride_comment, sCB_comment, sFinal_code]
            if sFinal_name_short != self.__sSql_IndividualMeasureMacro_tableAttributeUsed4sorting:
                iGold_total += 1
                if sGold in (None, ""):
                    iGold_unset += 1
                lstPANGAEA_meta_log.append(lstEntry_log)

            # DEV: Header im alternativen Pangaea style
            sName_pangaeaStyle = createPANGAEA(sFinal_name_short, sFinal_code, sFinal_comment, sFinal_unit)

            # Normale Header
            lstPANGAEA_data_header.append(sFinal_name_short)

        # *** Check Mapping: Header SQL Daten -> interne Datenstruktur (Name.short)
        if len(lstPANGAEA_data_header_sql) != len(lstPANGAEA_data_header):
            res.addErr("Mapping problem", add="Differing numbers of header", kat="Write EXCEL", sys="PANGAEA Exporter")
        lstErrLog = []
        for iIdx in range(0, len(lstPANGAEA_data_header)):
            if lstPANGAEA_data_header[iIdx] != lstPANGAEA_data_header_sql[iIdx]:
                lstErrLog.append(lstPANGAEA_data_header[iIdx] + "!=" + lstPANGAEA_data_header_sql[iIdx])
        if len(lstErrLog) > 0:
            res.addErr("Mapping problem", add="Differing header: " + PrettyText.lst2strC(lstErrLog), kat="Write EXCEL",
                       sys="PANGAEA Exporter")
            return res

        sLog += "--- Mapping table (total) ---\n\n"
        sLog += PrettyText.table(lstPANGAEA_meta_log)
        sLog += "\n\n"
        sLog += f'Unchecked entries: {iGold_unset}/{iGold_total} ({int(iGold_unset / iGold_total * 100)}%)'
        sLog += "\n\n"

        # *** Analysiere SQL Daten und erzeuge interne Datenstruktur
        # a) Entferne leere Zeilen
        # b) merke leere Spalten
        # c) merke ob Indivvidual Measure benutzt wird in einem Subset (... dann Individual Label mit ausgeben)
        workflow(top="Converter CB->PANGAEA", sub="Analysing data", fmt=False, status="Analyse data")
        iNrMax = len(lstPANGAEA_data_sql)
        setColEmpty = set()
        bIndMeasureMacro_empty = True
        setIndmeasureMacro_tmp = set()
        setIndmeasureMacro_used = set()
        bIndMeasureMeio_empty = True
        setIndmeasureMeio_tmp = set()
        setIndmeasureMeio_used = set()
        for iDx in range(0, len(lstPANGAEA_data_header_sql)):  # Init: Alle Columns sind leer
            setColEmpty.add(iDx)
        for iRow_Nr, tpRow in enumerate(lstPANGAEA_data_sql):
            bEmptyRow = True
            # Hier wird geprüft ob ein Individual Label (Biota Name) mehr als einmal pro Subset vorkommt
            # Sprich: ein IndMeasure ist anhängig an ein Biota
            # Erzeuge einmaliges Label für Biota/Individual Measure
            # Meio
            sEventIndmeasureMeioLabel_uniqId = creatEventImeasureId(tpRow)
            if sEventIndmeasureMeioLabel_uniqId is not None:
                if sEventIndmeasureMeioLabel_uniqId in setIndmeasureMacro_tmp:
                    setIndmeasureMeio_used.add(sEventIndmeasureMeioLabel_uniqId)
                else:
                    setIndmeasureMeio_tmp.add(sEventIndmeasureMeioLabel_uniqId)
            # Macro
            sEventIndmeasureMacroLabel_uniqId = creatEventImeasureId(tpRow)
            if sEventIndmeasureMacroLabel_uniqId is not None:
                if sEventIndmeasureMacroLabel_uniqId in setIndmeasureMacro_tmp:
                    setIndmeasureMacro_used.add(sEventIndmeasureMacroLabel_uniqId)
                else:
                    setIndmeasureMacro_tmp.add(sEventIndmeasureMacroLabel_uniqId)
            # Loop cols
            for iCol_nr, val in enumerate(tpRow):
                if bEmptyRow and (val not in (None, "")):  # Teste auf leere Zeile
                    bEmptyRow = False
                if (iCol_nr in setColEmpty) and (val not in (None, "")):
                    setColEmpty.remove(iCol_nr)  # Eintrag in Zeile wurde gefunden
                # Hier wird analysiert ob ein Eintrag in Individual Measure exisitert oder nicht?
                if bIndMeasureMeio_empty and (val not in (None, "")) and self.map_getPart(iCol_nr) == "leaf-indmeasure-meio" and (
                        self.map_getName(iCol_nr) not in ("Individual label")):
                    bIndMeasureMeio_empty = False
                if bIndMeasureMacro_empty and (val not in (None, "")) and self.map_getPart(iCol_nr) == "leaf-indmeasure" and (
                        self.map_getName(iCol_nr) not in ("Individual label")):
                    bIndMeasureMacro_empty = False
            if not bEmptyRow:
                # Nur volle Zeilen merken
                lstRow = list(tpRow)
                lstPANGAEA_data.append(lstRow)
            else:
                pass  # Darf eigentlich nicht passieren
            progress(iRow_Nr + 1, iNrMax, 100, proz=False, abs=True, cli=True, status="Analyse data")

        # *** Anpassungen für Anzeige von Individual Measure
        # Kein Individual Measure Eintrag vorhanden -> Definere "Individual label" als leere Column
        if bIndMeasureMacro_empty and bIndMeasureMeio_empty:
            setColEmpty.add(self.map_getIndex(self.__sSql_IndividualLabel_name))
        # Entferne immer die "indmea_id" Spalte (die wurde zwingend zum Sortieren von IndMeasure innerhalb von SQL benoetig)
        setColEmpty.add(self.map_getIndex(self.__sSql_IndividualMeasureMeio_tableAttributeUsed4sorting))  # Entferne immer die "indmea_meio_id" Spalte
        setColEmpty.add(self.map_getIndex(self.__sSql_IndividualMeasureMacro_tableAttributeUsed4sorting))  # Entferne immer die "indmea_macro_id" Spalte

        # *** Entferne leere Spalten in Headern (Data und Metadata)
        for iCol_Nr, sCol in enumerate(lstPANGAEA_data_header):
            if iCol_Nr in setColEmpty:
                pass
                # pv(f"delete empty col in data header: {sCol}")
            else:
                lstPANGAEA_data_header_clean.append(sCol)  # Entferne Header in Data
        for iCol_Nr, lstRow in enumerate(lstPANGAEA_meta):
            if iCol_Nr in setColEmpty:
                pass
                # print(f"delete empty row in meta: {iCol_Nr}")
            else:
                lstPANGAEA_meta_clean.append(lstRow)

        # *** Erzeuge Multisheet
        # a) entferne leere Spalten
        # b) blende Taxa Labels aus, wenn sie nicht benötige werden (Kein IndMeasure hängt dran)
        workflow(top="Converter CB->PANGAEA", sub="Remove empty cols and improve readability", fmt=False)
        ms = MultiSheet(name="Subset.PANGAEA")

        # --- data
        sLog += "--- Cleaning up and improving readability ---\n"
        if not bIndMeasureMacro_empty:
            sLog += "* Individual Measure data found: Show Individual labels\n"
        if len(setColEmpty) > 0:
            sLog += "* Removed empty cols:\n"
        for i in setColEmpty:
            sLog += f"{i}: {self.map_getName(i)}\n"
        for iRow_Nr, lstRow in enumerate(lstPANGAEA_data):
            progress(iRow_Nr + 1, iNrMax, 100, proz=False, abs=True, cli=True,
                     status="Remove empty rows and cols and improve readability")
            # Entferne hier die leeren Spalten
            lstRow_clean = []
            # Ausblenden von nötigen benutzten Taxa Label (kein Indivdual Measure hängt dran)
            if creatEventImeasureId(lstRow) not in setIndmeasureMacro_used:
                lstRow[self.map_getIndex(self.__sSql_IndividualLabel_name)] = ""
            for iCol_Nr, sVal in enumerate(lstRow):
                if iCol_Nr not in setColEmpty:
                    lstRow_clean.append(sVal)
            lstPANGAEA_data_clean.append(lstRow_clean)
        # header
        ms.addSheet(self.__sExcel_sheetname_data, lstPANGAEA_data_header_clean,
                    lstPANGAEA_data_clean)  # benutze normale Header

        # --- meta
        ms.addSheet(self.__sExcel_sheetname_meta, lstPANGAEA_meta_header, lstPANGAEA_meta_clean)

        # DEV
        # print(ms.getInfo(full=True))
        # print(res)
        # quit()

        # Antwort
        res.addData("multisheet", ms)
        res.addData("sql", sPANGAEA_Sql)
        res.addData("log", sLog)
        return res

    def convert_export_create_excel_wb(self, **kwargs):

        # Helper: Style whole row
        def helper_styleRow(lstRow, ws, style):
            lstCells = []
            if lstRow is not None:
                for sVal in lstRow:
                    cell = openpyxl.cell.WriteOnlyCell(ws, value=sVal)
                    cell.font = style
                    lstCells.append(cell)
            return lstCells

        workflow = self.__smartprinter.workflow
        progress = self.__smartprinter.progress

        res = Result("PANGAEA Export")

        # Parameter
        wb = kwargs.get("wb")
        ms = kwargs.get("multisheet")
        if ms is None:
            ms = MultiSheet()

        workflow(top="Converter CB->PANGAEA", sub="Create EXCEL workbook", fmt=False)

        # Workbook & Worksheets
        if wb is None:
            wb = openpyxl.Workbook()
            bFirstSheet = True
        else:
            bFirstSheet = False
        ws_data = wb.active  # first Sheet
        ws_data.title = self.__sExcel_sheetname_data
        ws_meta = wb.create_sheet(title=self.__sExcel_sheetname_meta)

        try:
            lstPANGAEA_data_header = ms.getColNames(self.__sExcel_sheetname_data)
        except:
            lstPANGAEA_data_header = []
        try:
            lstPANGAEA_meta_header = ms.getColNames(self.__sExcel_sheetname_meta)
        except:
            lstPANGAEA_meta_header = []

        # Write PANGAEA EXCEL

        # data
        ms.setOpenSheetName(self.__sExcel_sheetname_data)
        ws_data.append(helper_styleRow(lstPANGAEA_data_header, ws_data,
                                       openpyxl.styles.Font(italic=False, bold=True, underline=None)))  # header
        iRow = 1
        iRowMax = ms.getNoOfRows()
        for tpRow in ms():
            progress(iRow + 1, iRowMax, 100, proz=False, abs=True, cli=True, status="Create EXCEL data")
            ws_data.append(tpRow)
            iRow += 1

        # meta
        ms.setOpenSheetName(self.__sExcel_sheetname_meta)
        ws_meta.append(helper_styleRow(lstPANGAEA_meta_header, ws_meta,
                                       openpyxl.styles.Font(italic=False, bold=True, underline=None)))  # header
        for tpRow in ms():
            ws_meta.append(tpRow)

        res.addData("wb", wb)
        return res
